@@ -28,6 +28,7 @@ except ImportError:
 
 import base64
 import mimetypes
+import sys
 import time
 import urllib
 import zipfile
@@ -83,6 +84,9 @@ META = 'meta.xml'
 THUMBNAILS = 'Thumbnails'
 output_encode = genshi.output.encode
 EtreeElement = lxml.etree.Element
+XML_INVALID_CHAR_EXPR = re.compile(
+    # from https://www.w3.org/TR/REC-xml/#charsets
+    u'[\x00-\x08\x0b\x0c\x0e-\x1F\uD800-\uDFFF\uFFFE\uFFFF]')
 
 # A note regarding OpenDocument namespaces:
 #
@@ -111,9 +115,8 @@ class OOTemplateError(genshi.template.base.TemplateSyntaxError):
 class ImageHref:
     "A class used to add images in the odf zipfile"
 
-    def __init__(self, zfile, manifest, context):
-        self.zip = zfile
-        self.manifest = manifest
+    def __init__(self, serializer, context):
+        self.serializer = serializer
         self.context = context.copy()
 
     def __call__(self, expr):
@@ -129,9 +132,7 @@ class ImageHref:
         name = md5(file_content).hexdigest()
         path = 'Pictures/%s%s' % (
             name, mimetypes.guess_extension(mimetype))
-        if path not in self.zip.namelist():
-            self.zip.writestr(path, file_content)
-            self.manifest.add_file_entry(path, mimetype)
+        self.serializer.add_file(path, file_content, mimetype)
         return {'{http://www.w3.org/1999/xlink}href': path}
 
 
@@ -245,6 +246,14 @@ def update_py_attrs(node, value):
                 (node.attrib[py_attrs_attr], value)
 
 
+def escape_xml_invalid_chars(value, repl=' '):
+    "Replace invalid characters for XML."
+    if isinstance(value, basestring):
+        return XML_INVALID_CHAR_EXPR.sub(repl, value)
+    else:
+        return value
+
+
 class Template(MarkupTemplate):
 
     def __init__(self, source, filepath=None, filename=None, loader=None,
@@ -253,6 +262,7 @@ class Template(MarkupTemplate):
         self.inner_docs = []
         self.has_col_loop = False
         self._source = None
+        self._files = set()
         super(Template, self).__init__(source, filepath, filename, loader,
                                        encoding, lookup, allow_exec)
 
@@ -301,6 +311,7 @@ class Template(MarkupTemplate):
 
         parsed = []
         for fpath, fparsed in content_files + styles_files:
+            self._files.add(fpath)
             parsed.append((genshi.core.PI, ('relatorio', fpath), None))
             parsed += fparsed
 
@@ -397,20 +408,24 @@ class Template(MarkupTemplate):
                            % (self.filepath, expr)
             elif expr != statement.text and statement.tag == text_a:
                 warn_msg = "url and text do not match in %s: %s != %s" \
-                           % (self.filepath, expr,
-                              statement.text.encode('utf-8'))
+                           % (self.filepath, expr, statement.text)
             if warn_msg:
                 if directive is not None and not is_opening:
                     warn_msg += " corresponding to opening tag '%s'" \
-                                % opened_tags[-1].text
+                                % opened_tags[-1][0].text
                 warnings.warn(warn_msg)
 
             if directive in GENSHI_CLOSING_DIRECTIVE:
                 # map closing tags with their opening tag
                 if is_opening:
-                    opened_tags.append(statement)
+                    opened_tags.append((statement, directive))
                 else:
-                    closing_tags[id(opened_tags.pop())] = statement
+                    opening_statement, opening_directive = opened_tags.pop()
+                    assert directive == opening_directive, (
+                        "Wrong pairing tags between '%s' and '%s'"
+                        % (opening_statement.text.encode('utf-8'),
+                           statement.text.encode('utf-8')))
+                    closing_tags[id(opening_statement)] = statement
             # - we only need to return opening statements
             if is_opening:
                 r_statements.append((statement,
@@ -523,10 +538,15 @@ class Template(MarkupTemplate):
                 grand_parent = parent.getparent()
                 # Guess type only if it is the only value in the cell
                 # and its parent has no style
-                if (grand_parent is None
-                        or grand_parent.tag != table_cell_tag
-                        ) or len(grand_parent) != 1 or has_style(parent):
-                    r_node.attrib[py_replace] = expr
+                if ((grand_parent is None
+                            or grand_parent.tag != table_cell_tag)
+                        or len(grand_parent) != 1
+                        or len(parent) != 1
+                        or (parent.text and parent.text.strip())
+                        or (r_node.tail and r_node.tail.strip())
+                        or has_style(parent)):
+                    r_node.attrib[py_replace] = (
+                        '__relatorio_escape_invalid_chars(%s)' % expr)
                     continue
 
                 cache_id = id(r_node)
@@ -790,6 +810,7 @@ class Template(MarkupTemplate):
             type_ = 'float'
         elif isinstance(val, basestring):
             type_ = 'string'
+            val = escape_xml_invalid_chars(val)
         elif isinstance(val, datetime.timedelta):
             type_ = 'time'
             val = 'P%sD%sS' % (val.days, val.seconds)
@@ -804,12 +825,11 @@ class Template(MarkupTemplate):
 
     def generate(self, *args, **kwargs):
         "creates the RelatorioStream."
-        serializer = OOSerializer(self._source)
-        kwargs['__relatorio_make_href'] = ImageHref(serializer.outzip,
-                                                    serializer.manifest,
-                                                    kwargs)
+        serializer = OOSerializer(self._source, self._files)
+        kwargs['__relatorio_make_href'] = ImageHref(serializer, kwargs)
         kwargs['__relatorio_make_dimension'] = ImageDimension(self.namespaces)
         kwargs['__relatorio_guess_type'] = self._guess_type
+        kwargs['__relatorio_escape_invalid_chars'] = escape_xml_invalid_chars
 
         counter = ColumnCounter()
         kwargs['__relatorio_reset_col_count'] = counter.reset
@@ -829,6 +849,7 @@ class Template(MarkupTemplate):
             col_filter = Transformer('//repeat[namespace-uri()="%s"]'
                                      % RELATORIO_URI)
             col_filter = col_filter.apply(transformation)
+            # Must consume the stream to fill counter
             stream = Stream(list(stream), self.serializer) | col_filter
         return RelatorioStream(stream, serializer)
 
@@ -874,7 +895,8 @@ def get_zip_file(source):
 def fod2od(source):
     "Convert Flat OpenDocument to OpenDocument"
     odt_io = BytesIO()
-    odt_zip = zipfile.ZipFile(odt_io, mode='w')
+    odt_zip = zipfile.ZipFile(
+        odt_io, mode='w', compression=zipfile.ZIP_DEFLATED)
     fodt_tree = lxml.etree.parse(source)
     fodt_root = fodt_tree.getroot()
     office_ns = fodt_root.nsmap['office']
@@ -889,6 +911,8 @@ def fod2od(source):
         '{%s}body' % office_ns: ['content'],
         }
     mimetype = fodt_root.attrib['{%s}mimetype' % office_ns]
+    # mimetype should be written first to let detection through 'magic number'
+    odt_zip.writestr('mimetype', mimetype, zipfile.ZIP_STORED)
     documents = {}
     images = []
     for child in fodt_root:
@@ -917,7 +941,6 @@ def fod2od(source):
         odt_zip.writestr(fname, data)
         manifest.add_file_entry(fname, mime_type)
     odt_zip.writestr(MANIFEST, str(manifest))
-    odt_zip.writestr('mimetype', mimetype)
     odt_zip.close()
     return odt_io
 
@@ -931,7 +954,11 @@ def extract_images(child, namespaces, start=0):
         binary_data, = image.xpath(
             './office:binary-data', namespaces=namespaces)
         data = base64.b64decode(binary_data.text)
-        mime_type = magic.from_buffer(data, mime=True)
+        if hasattr(magic, 'from_buffer'):
+            mime_type = magic.from_buffer(data, mime=True)
+        else:
+            # Not python-magic but file-magic
+            mime_type = magic.detect_from_content(data).mime_type
         name = 'Pictures/image%s%s' % (
             i, mimetypes.guess_extension(mime_type))
         image.remove(binary_data)
@@ -1020,39 +1047,115 @@ class Meta(object):
         return val
 
 
+class _AbstractZipWriteSplitStream(object):
+    def __init__(self, zipfile, chunksize=64):
+        self.zipfile = zipfile
+        self.chunksize = chunksize
+
+    def open(self, zinfo):
+        raise NotImplementedError
+
+    def close(self):
+        raise NotImplementedError
+
+    def __call__(self, stream):
+        for kind, data, pos in stream:
+            if kind == genshi.core.PI and data[0] == 'relatorio':
+                self.open(data[1])
+                continue
+            yield kind, data, pos
+        self.close()
+
+    def write(self, data):
+        raise NotImplementedError
+
+
+if sys.version_info >= (3, 6):
+    class _ZipWriteSplitStream(_AbstractZipWriteSplitStream):
+        def __init__(self, *args, **kwargs):
+            super(_ZipWriteSplitStream, self).__init__(*args, **kwargs)
+            self._fp = None
+            self._buffer = []
+            self._zinfo = None
+
+        def open(self, zinfo):
+            if self._fp or self._buffer:
+                self.close()
+            self._zinfo = zinfo
+            self._fp = None
+
+        def close(self):
+            self.flush()
+            self._fp.close()
+            self._zinfo = None
+            self._fp = None
+
+        def write(self, data):
+            self._buffer.append(data)
+            if len(self._buffer) > self.chunksize:
+                self.flush()
+
+        def flush(self):
+            if not self._fp:
+                self._fp = self.zipfile.open(self._zinfo, mode='w')
+            self._fp.write(b''.join(self._buffer))
+            self._buffer.clear()
+else:
+    class _ZipWriteSplitStream(_AbstractZipWriteSplitStream):
+        def __init__(self, *args, **kwargs):
+            super(_ZipWriteSplitStream, self).__init__(*args, **kwargs)
+            self._fp = None
+            self._zinfo = None
+
+        def open(self, zinfo):
+            if self._fp:
+                self.close()
+            self._zinfo = zinfo
+            self._fp = BytesIO()
+
+        def close(self):
+            self._fp.seek(0)
+            self.zipfile.writestr(self._zinfo, self._fp.read())
+            self._zinfo = None
+            self._fp = None
+
+        def write(self, data):
+            self._fp.write(data)
+
+
 class OOSerializer:
 
-    def __init__(self, source):
+    def __init__(self, source, files, chunksize=64):
         self.inzip = get_zip_file(source)
         self.manifest = Manifest(self.inzip.read(MANIFEST))
         self.meta = Meta(self.inzip.read(META))
-        self.new_oo = BytesIO()
-        self.outzip = zipfile.ZipFile(self.new_oo, 'w')
         self.xml_serializer = genshi.output.XMLSerializer()
+        self._files = files
+        self.chunksize = chunksize
+        self.outzip = None
+        self._deferred = None
 
-    def __call__(self, stream):
+    def __call__(self, stream, method=None, encoding='utf-8', out=None):
+        if out is None:
+            result = BytesIO()
+        else:
+            result = out
+        self.outzip = zipfile.ZipFile(
+            result, mode='w', compression=zipfile.ZIP_DEFLATED)
+        self._deferred = []
         files = {}
-        for kind, data, pos in stream:
-            if kind == genshi.core.PI and data[0] == 'relatorio':
-                stream_for = data[1]
-                continue
-            files.setdefault(stream_for, []).append((kind, data, pos))
-
         now = time.localtime()[:6]
         manifest_info = None
         for f_info in self.inzip.infolist():
             if f_info.filename.startswith('ObjectReplacements'):
                 continue
-            elif f_info.filename in files:
-                stream = files[f_info.filename]
+            elif f_info.filename in self._files:
                 # create a new file descriptor, copying some attributes from
                 # the original file
                 new_info = zipfile.ZipInfo(f_info.filename, now)
                 for attr in ('compress_type', 'flag_bits', 'create_system'):
                     setattr(new_info, attr, getattr(f_info, attr))
-                serialized_stream = output_encode(self.xml_serializer(stream),
-                    encoding='utf-8')
-                self.outzip.writestr(new_info, serialized_stream)
+                files[f_info.filename] = new_info
             elif f_info.filename == MANIFEST:
                 manifest_info = f_info
             elif f_info.filename == META:
@@ -1061,12 +1164,29 @@ class OOSerializer:
                 self.manifest.remove_file_entry(f_info.filename)
             else:
                 self.outzip.writestr(f_info, self.inzip.read(f_info.filename))
+
+        writer = _ZipWriteSplitStream(self.outzip, self.chunksize)
+        output_encode(
+            self.xml_serializer(writer(stream)), encoding=encoding, out=writer)
+
+        for args in self._deferred:
+            self.add_file(*args)
         self.manifest.remove_file_entry(THUMBNAILS + '/')
         if manifest_info:
             self.outzip.writestr(manifest_info, str(self.manifest))
         self.inzip.close()
         self.outzip.close()
 
-        return self.new_oo
+        if out is None:
+            return result
+
+    def add_file(self, path, content, mimetype):
+        if self.outzip and path not in self.outzip.namelist():
+            try:
+                self.outzip.writestr(path, content)
+                self.manifest.add_file_entry(path, mimetype)
+            except ValueError:
+                self._deferred.append((path, content, mimetype))
+
 
 MIMETemplateLoader.add_factory('oo.org', Template)
