@@ -31,7 +31,8 @@ from genshi.core import PI
 from genshi.template.eval import UndefinedError
 
 from relatorio.templates.opendocument import Template, GENSHI_EXPR,\
-    GENSHI_URI, RELATORIO_URI, fod2od, remove_node_keeping_tail
+    GENSHI_URI, RELATORIO_URI, fod2od, remove_node_keeping_tail, \
+    escape_xml_invalid_chars
 
 OO_TABLE_NS = "urn:oasis:names:tc:opendocument:xmlns:table:1.0"
 
@@ -100,7 +101,94 @@ class TestOOTemplating(unittest.TestCase):
         root_interpolated = lxml.etree.parse(interpolated).getroot()
         child = root_interpolated[0]
         self.assertEqual(
-            child.get('{http://genshi.edgewall.org/}replace'), 'foo')
+            child.get('{http://genshi.edgewall.org/}replace'),
+            '__relatorio_escape_invalid_chars(foo)')
+
+    @unittest.skipIf(
+        not hasattr(unittest.TestCase, 'assertRegex'),
+        'assertRegex not supported')
+    def test_directives_in_cell(self):
+        "Testing the guess type of directive inside cell"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:table="urn:table" xmlns:xlink="urn:xlink">
+                    <table:table-cell>
+                        <text:p>
+                            <text:a xlink:href="relatorio://foo">foo</text:a>
+                        </text:p>
+                    </table:table-cell>
+                 </xml>'''
+        interpolated = self.oot.insert_directives(xml)
+        root_interpolated = lxml.etree.parse(interpolated).getroot()
+        cell = root_interpolated[0]
+        self.assertRegex(
+            cell.get('{http://genshi.edgewall.org/}attrs'),
+            '__relatorio_guess_type\(__relatorio_store_cache\([0-9]*, foo\)\)')
+        child = cell[0][0]
+        self.assertRegex(
+            child.get('{http://genshi.edgewall.org/}replace'),
+            '__relatorio_get_cache\([0-9]*\)')
+
+    def test_directives_in_cell_with_children(self):
+        "Testing directive inside cell with children"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:table="urn:table" xmlns:xlink="urn:xlink">
+                    <table:table-cell>
+                        <text:p>
+                            <text:a xlink:href="relatorio://foo">foo</text:a>
+                        </text:p>
+                        <text:p>bar</text:p>
+                    </table:table-cell>
+                 </xml>'''
+        interpolated = self.oot.insert_directives(xml)
+        root_interpolated = lxml.etree.parse(interpolated).getroot()
+        cell = root_interpolated[0]
+        self.assertEqual(len(cell), 2)
+        self.assertFalse(
+            cell.get('{http://genshi.edgewall.org/}attrs'))
+        child = cell[0][0]
+        self.assertEqual(
+            child.get('{http://genshi.edgewall.org/}replace'),
+            '__relatorio_escape_invalid_chars(foo)')
+
+    def test_directives_in_cell_with_text(self):
+        "Testing directive inside cell with text"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:table="urn:table" xmlns:xlink="urn:xlink">
+                    <table:table-cell>
+                        <text:p>
+                            bar
+                            <text:a xlink:href="relatorio://foo">foo</text:a>
+                        </text:p>
+                    </table:table-cell>
+                 </xml>'''
+        interpolated = self.oot.insert_directives(xml)
+        root_interpolated = lxml.etree.parse(interpolated).getroot()
+        cell = root_interpolated[0]
+        self.assertEqual(len(cell), 1)
+        self.assertFalse(
+            cell.get('{http://genshi.edgewall.org/}attrs'))
+        child = cell[0][0]
+        self.assertEqual(
+            child.get('{http://genshi.edgewall.org/}replace'),
+            '__relatorio_escape_invalid_chars(foo)')
+
+    def test_directives_in_cell_with_tail(self):
+        "Testing directive inside cell with text"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:table="urn:table" xmlns:xlink="urn:xlink">
+                    <table:table-cell>
+                        <text:p>
+                            <text:a xlink:href="relatorio://foo">foo</text:a>
+                            bar
+                        </text:p>
+                    </table:table-cell>
+                 </xml>'''
+        interpolated = self.oot.insert_directives(xml)
+        root_interpolated = lxml.etree.parse(interpolated).getroot()
+        cell = root_interpolated[0]
+        self.assertEqual(len(cell), 1)
+        self.assertFalse(
+            cell.get('{http://genshi.edgewall.org/}attrs'))
+        child = cell[0][0]
+        self.assertEqual(
+            child.get('{http://genshi.edgewall.org/}replace'),
+            '__relatorio_escape_invalid_chars(foo)')
 
     def test_column_looping(self):
         xml = b'''
@@ -228,6 +316,49 @@ class TestOOTemplating(unittest.TestCase):
         self.assertTrue(last_row_node.get("value")
                         .startswith('__relatorio_store_col_count'))
 
+    @unittest.skipIf(
+        not getattr(unittest.TestCase, 'assertWarnsRegex', None),
+        "assertWarns not supported")
+    def test_statement_no_text_warning(self):
+        "Test warning for missing statement text"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
+                    <text:a xlink:href="relatorio://content:foo"></text:a>
+                  </xml>'''
+
+        with self.assertWarnsRegex(
+                UserWarning, r"No statement text in '.*' for 'content:foo'"):
+            self.oot.insert_directives(xml)
+
+    @unittest.skipIf(
+        not getattr(unittest.TestCase, 'assertWarnsRegex', None),
+        "assertWarns not supported")
+    def test_statement_missmatch_text_warning(self):
+        "Test warning for missing statement text"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
+                    <text:a xlink:href="relatorio://content:foo">content:bar</text:a>
+                  </xml>'''
+
+        with self.assertWarnsRegex(
+                UserWarning,
+                r"url and text do not match in .*: "
+                "content:foo != content:bar"):
+            self.oot.insert_directives(xml)
+
+    @unittest.skipIf(
+        not getattr(unittest.TestCase, 'assertWarnsRegex', None),
+        "assertWarns not supported")
+    def test_statement_text_warning_closing(self):
+        "Test warning for statement text in closing"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
+                    <text:a xlink:href="relatorio://with foo='test'">with foo='test'</text:a>
+                    <text:a xlink:href="relatorio:///with"></text:a>
+                  </xml>'''
+
+        with self.assertWarnsRegex(
+                UserWarning,
+                r".* corresponding to opening tag 'with foo='test'"):
+            self.oot.insert_directives(xml)
+
     def test_text_outside_p(self):
         "Testing that the tail text of a directive node is handled properly"
         xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
@@ -412,3 +543,23 @@ class TestRemoveNodeKeepingTail(unittest.TestCase):
         self.assertEqual(
             lxml.etree.tostring(tree),
             b'''<parent><previous/>tailtail</parent>''')
+
+
+class TestEscapeXMLInvalidChars(unittest.TestCase):
+
+    def test_valid(self):
+        "Test escape valid"
+        self.assertEqual(escape_xml_invalid_chars("foo"), "foo")
+
+    def test_valid_control(self):
+        "Test escape valid control"
+        self.assertEqual(escape_xml_invalid_chars("foo \x09"), "foo \x09")
+
+    def test_invalid(self):
+        "Test escape invalid"
+        self.assertEqual(escape_xml_invalid_chars("foo \x00 bar"), "foo   bar")
+
+    def test_replacement(self):
+        "Test replacement"
+        self.assertEqual(
+            escape_xml_invalid_chars("foo \x00 bar", "?"), "foo ? bar")
